/*     %W% %G%	*/

/**********************************************************************
FILE   : %M%
PURPOSE: SNNS-Kernel Learning Functions
NOTES  : with following learning functions:
	   - Backpropagation
	   - Backpropagation with momentum term
           - Quickprop
	   - Counterpropagation
	   - BackPercolation

AUTHOR : Niels Mache
DATE   : 01.10.90
VERSION : %I%  %G%

 Copyright (c) 1990,1991,1992 by Niels Mache and the SNNS Group

**********************************************************************/

#include <stdio.h>
#include <math.h>
#include <values.h>

#include "kr_typ.h"	 /*  Kernel Types and Constants  */
#include "kr_const.h"	 /*  Constant Declarators for SNNS-Kernel  */
#include "kr_def.h"	 /*  Default Values  */
#include "kernel.h"	 /*  kernel function prototypes  */
#include "kr_mac.h"	 /*  Kernel Macros   */


#define  NET_ERROR( param )      param[ 0 ]  /*    returns the net error   */
#define  LEARN_PARAM1( param )   param[ 0 ]  /*    contains the 1st learning parameter  */
#define  LEARN_PARAM2( param )   param[ 1 ]  /*    contains the 2nd learning parameter  */
#define  LEARN_PARAM3( param )   param[ 2 ]  /*    contains the 3rd learning parameter  */
#define  LEARN_PARAM4( param )   param[ 3 ]  /*    contains the 4th learning parameter  */
#define  LEARN_PARAM5( param )   param[ 4 ]  /*    contains the 5th learning parameter  */

#define  SIGMOID_PRIME_OFFSET    0.1         /*    for modificated sigmoid function    */
#define SQR(x)  (x) * (x)


/*#################################################

GROUP: Global Var's

#################################################*/

extern bool    NetModified,	 /*  TRUE, if the network topology was modified  */
	       PatternsValid,	 /*  TRUE, if In-Core patterns are valid  */
               NetInitialize,    /*  TRUE, if the network has been initialized  */
               LearnFuncHasChanged;  /*  TRUE, if the learning function has been changed  */

extern int  NoOfUnits,		 /*  no. of units in the network  */
	    MinUnitNo,		 /*  the first (lowest) used unit number in the network  */
	    MaxUnitNo,		 /*  the last (highest) used unit number in the network  */
	    NoOfPatternPairs,	 /*  no. of defined pattern pairs */
	    NoOfShuffledPatterns,  /*  no. of entries in the pattern_number array  */
	    NoOfInputUnits,	 /*  no. of input units  */
            NoOfOutputUnits,     /*  no. of output units  */
	    NoOfHiddenUnits,	 /*  no. of hidden units  */
	    TopoSortID; 	 /*  topologic mode identifier	*/

extern UnitArray       unit_array;  /*	the unit array	*/

extern TopoPtrArray    topo_ptr_array;	/*  stores pointers to topological sorted units
					    used by kr_topoSort()  */

extern Patterns        in_patterns,	/*  the input pattern array  */
		       out_patterns;	/*  the output pattern array  */
extern PatternNumbers  pattern_numbers;   /*  contains shuffled pattern numbers
					      used by kr_shufflePatterns()  */

extern int     no_of_topo_units;	/*  no. of unit pointers in the topo_ptr_array	*/




/*#################################################

GROUP: Local Vars

#################################################*/

static int  NoOfLearnedPatterns;



/*  --------------   Learning  Functions  ----------------------    */



/*#################################################

GROUP: backpropagation learning function

#################################################*/


/*  backprop initialisation
*/
static krui_err initializeBackprop()
{
  register struct Unit   *unit_ptr;

  /*  set unit's bias to zero  */

  FOR_ALL_UNITS( unit_ptr )
    if ( (unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE)
      /*  unit is in use  */
      unit_ptr->bias = (FlintType) 0;

  return( KRERR_NO_ERROR );
}



/*  topological forward propagation
*/
static	void  propagateNetForward( pattern_no )
int     pattern_no;
{
  register struct Unit   *unit_ptr;
  register Patterns  in_pat;
  register TopoPtrArray     topo_ptr;


  /*  calculate startaddress for input pattern array  */
  in_pat = in_patterns + pattern_no * NoOfInputUnits;

  topo_ptr = topo_ptr_array;

  /*  copy pattern into input unit's activation and
      calculate output of the input units
  */
  while ((unit_ptr = *++topo_ptr) != NULL)
    {  /*  topo_ptr points to a (topological sorted) unit stucture (input units first)  */
    if (unit_ptr->out_func == OUT_IDENTITY)
      /*  identity output function: there is no need to call the output function  */
      unit_ptr->Out.output = unit_ptr->act = *in_pat++;
    else
      /*  no identity output function: calculate unit's output also  */
      unit_ptr->Out.output = (*unit_ptr->out_func) (unit_ptr->act = *in_pat++);
    }

  /*  popagate hidden units  */
  while ((unit_ptr = *++topo_ptr) != NULL)
    {  /*  topo_ptr points to a (topological sorted) unit stucture */
    /*  clear error values  */
    unit_ptr->Aux.flint_no = 0.0;

    /*  calculate the activation value of the unit: call the activation function if needed  */
    unit_ptr->act = (*unit_ptr->act_func) (unit_ptr);

    if (unit_ptr->out_func == OUT_IDENTITY)
      /*  identity output function: there is no need to call the output function  */
      unit_ptr->Out.output = unit_ptr->act;
    else
      /*  no identity output function: calculate unit's output also  */
      unit_ptr->Out.output = (*unit_ptr->out_func) (unit_ptr->act);
    }

  /*  popagate output units  */
  while ((unit_ptr = *++topo_ptr) != NULL)
    {  /*  topo_ptr points to a (topological sorted) unit stucture */
    /*  clear error values  */
    unit_ptr->Aux.flint_no = 0.0;

    /*  calculate the activation value of the unit: call the activation function if needed  */
    unit_ptr->act = (*unit_ptr->act_func) (unit_ptr);

    if (unit_ptr->out_func == OUT_IDENTITY)
      /*  identity output function: there is no need to call the output function  */
      unit_ptr->Out.output = unit_ptr->act;
    else
      /*  no identity output function: calculate unit's output also  */
      unit_ptr->Out.output = (*unit_ptr->out_func) (unit_ptr->act);
    }
}


/*  Backward error propagation (topological)
*/
static	float propagateNetBackward2( pattern_no, learn_parameter, delta_max )
int     pattern_no;
float   learn_parameter, delta_max;
{
  register struct Link   *link_ptr;
  register struct Site   *site_ptr;
  register struct Unit   *unit_ptr;
  register Patterns      out_pat;
  register float         error,  sum_error,  eta,  devit, learn_error;
  register TopoPtrArray     topo_ptr;


  sum_error = 0.0;    /*  reset network error  */
  eta = learn_parameter;   /*  store learn_parameter in CPU register  */
  /*  calculate address of the output pattern (with number pattern_no + 1)  */
  out_pat = out_patterns + (pattern_no + 1) * NoOfOutputUnits;

  /*  add 3 to no_of_topo_units because the topologic array contains
      4 NULL pointers  */
  topo_ptr = topo_ptr_array + (no_of_topo_units + 3);

  /*  calculate output units only  */
  while ((unit_ptr = *--topo_ptr) != NULL)
    {
    devit = *(--out_pat) - unit_ptr->Out.output;  /*  calc. devitation	*/
    if ( (float) fabs( devit ) <= delta_max )  continue;

    sum_error += devit * devit;  /*  sum up the error of the network  */

    /*	calc. error for output units	 */
    error = devit * (unit_ptr->act_deriv_func) ( unit_ptr );
    /*	calc. the error for adjusting weights and bias of the predecessor units  */
    learn_error = eta * error;
    /*	adjust bias value  */
    unit_ptr->bias += learn_error;

    if (UNIT_HAS_DIRECT_INPUTS( unit_ptr ))
      {  /*  the unit has direkt links  */
      FOR_ALL_LINKS( unit_ptr, link_ptr )
	{  /*  adjust link weights and calc. sum of errors of the predecessor units  */
        link_ptr->to->Aux.flint_no += link_ptr->weight * error;
        link_ptr->weight += learn_error * link_ptr->to->Out.output;
        }
      }
    else
      {  /*	the unit has sites  */
      FOR_ALL_SITES_AND_LINKS( unit_ptr, site_ptr, link_ptr )
        {  /*  adjust link weights and calc. sum of errors of the predecessor units */
        link_ptr->to->Aux.flint_no += link_ptr->weight * error;
        link_ptr->weight += learn_error * link_ptr->to->Out.output;
        }
      }
    }

  /*  calculate hidden units only  */
  while ((unit_ptr = *--topo_ptr) != NULL)
    {
    /*	calc. the error of the (hidden) unit  */
    error = (unit_ptr->act_deriv_func) ( unit_ptr ) * unit_ptr->Aux.flint_no;
    /*	calc. the error for adjusting weights and bias of the predecessor units  */
    learn_error = eta * error;
    /*	adjust bias value  */
    unit_ptr->bias += learn_error;

    if (UNIT_HAS_DIRECT_INPUTS( unit_ptr ))
      {  /*  the unit has direkt links	*/
      FOR_ALL_LINKS( unit_ptr, link_ptr )
	{  /*  adjust link weights and calc. sum of errors of the predecessor units  */
	if (link_ptr->to->flags & UFLAG_TTYP_HIDD)
	  /*  this link points to a hidden unit: sum up the error's from previos units  */
	  link_ptr->to->Aux.flint_no += link_ptr->weight * error;

	link_ptr->weight += learn_error * link_ptr->to->Out.output;
	}
      }
    else
      {  /*  the unit has sites  */
      FOR_ALL_SITES_AND_LINKS( unit_ptr, site_ptr, link_ptr )
	{  /*  adjust link weights and calc. sum of errors of the predecessor units */
	if (link_ptr->to->flags & UFLAG_TTYP_HIDD)
	  /*  this link points to a hidden unit: sum up the error's from previos units  */
	  link_ptr->to->Aux.flint_no += link_ptr->weight * error;

	link_ptr->weight += learn_error * link_ptr->to->Out.output;
	}
      }
    }

  sum_error *= 0.5;
  return( sum_error );  /*  return the error of the network */
}


/*  Backpropagation Learning Function
    Input Parameters:   1 : learning parameter
                        2 : delta max

    Output Parameters:  1 : error of the network (sum of all cycles)

*/
krui_err  LEARN_backprop( start_pattern, end_pattern, parameterInArray, NoOfInParams, parameterOutArray, NoOfOutParams )
int     start_pattern, end_pattern;
float   parameterInArray[],
        * *parameterOutArray;
int  NoOfInParams,
     *NoOfOutParams;
{
  static  float  OutParameter[1];   /*	OutParameter[0] stores the learning error  */
  int   i, ret_code, pattern_no;


  if (NoOfUnits == 0)  return( KRERR_NO_UNITS );  /*  No Units defined	*/
  if (NoOfInParams < 1)          /*  ####  have to be changed (must be 2)  #### */
    return( KRERR_PARAMETERS );  /*  Not enough input parameters  */

  *NoOfOutParams = 1;   /*  One return value is available (the learning error)  */
  *parameterOutArray = OutParameter;   /*  set the output parameter reference  */ 
  ret_code = KRERR_NO_ERROR;  /*  reset return code  */

  if (NetModified || (TopoSortID != TOPOLOGICAL_FF))
    {  /*  Net has been modified or topologic array isn't initialized */
    /*  check the topology of the network  */
    ret_code = kr_topoCheck();
    if (ret_code < KRERR_NO_ERROR)  return( ret_code );  /*  an error has occured  */
    if (ret_code < 2)  return( KRERR_NET_DEPTH );  /*  the network has less then 2 layers  */

    /*	count the no. of I/O units and check the patterns  */
    ret_code = kr_IOCheck();
    if (ret_code < KRERR_NO_ERROR)  return( ret_code );

    /*	sort units by topology and by topologic type  */
    ret_code = kr_topoSort( TOPOLOGICAL_FF );
    if ((ret_code != KRERR_NO_ERROR) && (ret_code != KRERR_DEAD_UNITS))
      return( ret_code );

    NetModified = FALSE;
    }

  if (NetInitialize || LearnFuncHasChanged)
    {  /*  Net has been modified or initialized, initialize backprop now  */
    ret_code = initializeBackprop();
    if (ret_code != KRERR_NO_ERROR)  return( ret_code );
    }

  NET_ERROR( OutParameter ) = 0.0;   /*  reset network error value  */
  for (i = start_pattern; i <= end_pattern; i++)
    {
    if (NoOfShuffledPatterns == NoOfPatternPairs)
      pattern_no = pattern_numbers[i];  /*  shuffle patterns   */
    else
      pattern_no = i;  /*  do not schuffle patterns  */

    propagateNetForward( pattern_no );	/*  Forward propagation */

    /*  Backward propagation  */
    /*  1st parameter is the learning parameter
	2nd parameter is the max. devitation between output pattern and
	the output of the output unit (delta max)
    */
    NET_ERROR( OutParameter ) +=
      propagateNetBackward2( pattern_no,
			     LEARN_PARAM1( parameterInArray ),
			     LEARN_PARAM2( parameterInArray ) );
    }

  return( ret_code );
}


/*  ----------  End of backpropagation learning function ----------  */


/*  -----------------  Begin of new learning function -----------------  */


/*#################################################

GROUP: backpropagation learning function with momentum term

#################################################*/

/*  backprop-momentum initialisation
*/
static krui_err initializeBackpropMomentum()
{
  register FlagWord      flags;
  register struct Link   *link_ptr;
  register struct Unit   *unit_ptr;
  register struct Site   *site_ptr;


  FOR_ALL_UNITS( unit_ptr )
    {
    flags = unit_ptr->flags;

    if ( (flags & UFLAG_IN_USE) == UFLAG_IN_USE)
      {  /*  unit is in use  */
      unit_ptr->bias = unit_ptr->value_a = (FlintType) 0;

      if (flags & UFLAG_SITES)
	{  /*  unit has sites  */
        FOR_ALL_SITES_AND_LINKS( unit_ptr, site_ptr, link_ptr )
          link_ptr->value_b = (FlintType) 0;
        }
      else
	{  /*  unit has no sites   */
	if (flags & UFLAG_DLINKS)
	  {  /*  unit has direct links	 */
          FOR_ALL_LINKS( unit_ptr, link_ptr )
            link_ptr->value_b = (FlintType) 0;
          }
        }
      }
    }

  return( KRERR_NO_ERROR );
}



/*  Backward error propagation (topological) of backpropagation learnig function
    with momentum term and flat spot elimination
*/
static	float Backprop_momentum_FSE( pattern_no, learn_parameter, mu, FSE_term, delta_max )
int     pattern_no;
float   learn_parameter, mu, FSE_term, delta_max;
{
  register struct Link   *link_ptr;
  register struct Site   *site_ptr;
  register struct Unit   *unit_ptr;
  register Patterns      out_pat;
  register float         error,  sum_error,  eta,  devit, learn_error;
  register TopoPtrArray     topo_ptr;


  sum_error = 0.0;    /*  reset network error  */
  eta = learn_parameter;   /*  store learn_parameter in CPU register  */
  /*  calculate address of the output pattern (with number pattern_no + 1)  */
  out_pat = out_patterns + (pattern_no + 1) * NoOfOutputUnits;


  /*  add 3 to no_of_topo_units because the topologic array contains
      4 NULL pointers  */
  topo_ptr = topo_ptr_array + (no_of_topo_units + 3);

  /*  calculate output units only  */
  while ((unit_ptr = *--topo_ptr) != NULL)
    {
    devit = *(--out_pat) - unit_ptr->Out.output;  /*  calc. devitation	*/
    if ( (float) fabs( devit ) <= delta_max )  continue;

    sum_error += devit * devit;  /*  sum up the error of the network  */
    /*	calc. error for output units	 */
    error = devit * ((unit_ptr->act_deriv_func) ( unit_ptr ) + FSE_term);

    /*	calc. the error for adjusting weights and bias of the predecessor units  */
    learn_error = eta * error;

    unit_ptr->value_a = learn_error + mu * unit_ptr->value_a;
    /*  adjust bias value  */
    unit_ptr->bias += unit_ptr->value_a;

    if (UNIT_HAS_DIRECT_INPUTS( unit_ptr ))
      {  /*  the unit has direkt links  */
      FOR_ALL_LINKS( unit_ptr, link_ptr )
	{  /*  adjust link weights and calc. sum of errors of the predecessor units  */
        link_ptr->to->Aux.flint_no += link_ptr->weight * error;
        link_ptr->weight += learn_error * link_ptr->to->Out.output;
        }
      }
    else
      {  /*	the unit has sites  */
      FOR_ALL_SITES_AND_LINKS( unit_ptr, site_ptr, link_ptr )
        {  /*  adjust link weights and calc. sum of errors of the predecessor units */
        link_ptr->to->Aux.flint_no += link_ptr->weight * error;
        link_ptr->weight += learn_error * link_ptr->to->Out.output;
        }
      }
    }


  /*  calculate hidden units only  */
  while ((unit_ptr = *--topo_ptr) != NULL)
    {
    /*	calc. the error of the (hidden) unit  */
    error = unit_ptr->Aux.flint_no * ((unit_ptr->act_deriv_func) ( unit_ptr ) + FSE_term);
   
    /*  calc. the error for adjusting weights and bias of the predecessor units	*/
    learn_error = eta * error;
    unit_ptr->value_a = learn_error + mu * unit_ptr->value_a;
    /*  adjust bias value  */
    unit_ptr->bias += unit_ptr->value_a;

    if (UNIT_HAS_DIRECT_INPUTS( unit_ptr ))
      {  /*  the unit has direkt links	*/
      FOR_ALL_LINKS( unit_ptr, link_ptr )
	{  /*  adjust link weights and calc. sum of errors of the predecessor units  */
	if (link_ptr->to->flags & UFLAG_TTYP_HIDD)
	  /*  this link points to a hidden unit: sum up the error's from previos units  */
	  link_ptr->to->Aux.flint_no += link_ptr->weight * error;

	link_ptr->value_b = learn_error * link_ptr->to->Out.output + mu * link_ptr->value_b;
	link_ptr->weight += link_ptr->value_b;
	}
      }
    else
      {  /*  the unit has sites  */
      FOR_ALL_SITES_AND_LINKS( unit_ptr, site_ptr, link_ptr )
	  {  /*  adjust link weights and calc. sum of errors of the predecessor units */
	  if (link_ptr->to->flags & UFLAG_TTYP_HIDD)
	    /*	this link points to a hidden unit: sum up the error's from previos units  */
	    link_ptr->to->Aux.flint_no += link_ptr->weight * error;

	  link_ptr->value_b = learn_error * link_ptr->to->Out.output + mu * link_ptr->value_b;
	  link_ptr->weight += link_ptr->value_b;
	  }
      }
    }

  sum_error *= 0.5;
  return( sum_error );  /*  return the error of the network */
}


/*  Backpropagation Learning Function
    Input Parameters:   1 : learning parameter
                        2 : momentum term
                        3 : flat-spot-elimination value
                        4 : delta max

    Output Parameters:  1 : error of the network (sum of all cycles)

*/
krui_err  LEARN_backpropMomentum( start_pattern, end_pattern, 
                                   parameterInArray, NoOfInParams, 
                                   parameterOutArray, NoOfOutParams )
int     start_pattern, end_pattern;
float   parameterInArray[],
        * *parameterOutArray;
int  NoOfInParams,
     *NoOfOutParams;
{
  static  float  OutParameter[1];   /*	OutParameter[0] stores the learning error  */
  int   i, ret_code, pattern_no;


  if (NoOfUnits == 0)  return( KRERR_NO_UNITS );  /*  No Units defined	*/
  if (NoOfInParams < 1)          /*  ####  have to be changed (must be 2)  #### */
    return( KRERR_PARAMETERS );  /*  Not enough input parameters  */

  *NoOfOutParams = 1;   /*  One return value is available (the learning error)  */
  *parameterOutArray = OutParameter;   /*  set the output parameter reference  */ 
  ret_code = KRERR_NO_ERROR;  /*  reset return code  */

  if (NetModified || (TopoSortID != TOPOLOGICAL_FF))
    {  /*  Net has been modified or topologic array isn't initialized */
    /*  check the topology of the network  */
    ret_code = kr_topoCheck();
    if (ret_code < KRERR_NO_ERROR)  return( ret_code );  /*  an error has occured  */
    if (ret_code < 2)  return( KRERR_NET_DEPTH );  /*  the network has less then 2 layers  */

    /*	count the no. of I/O units and check the patterns  */
    ret_code = kr_IOCheck();
    if (ret_code < KRERR_NO_ERROR)  return( ret_code );

    /*	sort units by topology and by topologic type  */
    ret_code = kr_topoSort( TOPOLOGICAL_FF );
    if ((ret_code != KRERR_NO_ERROR) && (ret_code != KRERR_DEAD_UNITS))
      return( ret_code );

    NetModified = FALSE;
    }

  if (NetInitialize || LearnFuncHasChanged)
    {  /*  Net has been modified or initialized, initialize backprop now  */
    ret_code = initializeBackpropMomentum();
    if (ret_code != KRERR_NO_ERROR)  return( ret_code );
    }

  NET_ERROR( OutParameter ) = 0.0;   /*  reset network error value  */
  for (i = start_pattern; i <= end_pattern; i++)
    {
    if (NoOfShuffledPatterns == NoOfPatternPairs)
      pattern_no = pattern_numbers[i];  /*  shuffle patterns   */
    else
      pattern_no = i;  /*  do not schuffle patterns  */

    propagateNetForward( pattern_no );	/*  Forward propagation */

    /*  Backward propagation  */
    /*  1st parameter is the learning parameter
        2nd parameter is the momentum term
        3rd parameter is the flat-spot-elimination value
        4th parameter is the max. devitation between output pattern and the output of the output unit
        (delta max)
    */
    NET_ERROR( OutParameter ) +=
      Backprop_momentum_FSE( pattern_no,
                             LEARN_PARAM1( parameterInArray ),
                             LEARN_PARAM2( parameterInArray ),
                             LEARN_PARAM3( parameterInArray ),
                             LEARN_PARAM4( parameterInArray ) );
    }

  return( ret_code );
}





/*#################################################

GROUP: quickpropagation learning function
       by Peter Zimmerer, modified by Niels Mache

#################################################*/



/*  quickprop initialisation
*/
static krui_err initializeQuickprop()
{
  register unsigned short    flags;
  register struct Link   *link_ptr;
  register struct Unit   *unit_ptr;
  register struct Site   *site_ptr;


  FOR_ALL_UNITS( unit_ptr )
    {
    flags = unit_ptr->flags;

    if ( (flags & UFLAG_IN_USE) == UFLAG_IN_USE)
      {  /*  unit is in use  */
      unit_ptr->bias = unit_ptr->value_a = unit_ptr->value_b =
      unit_ptr->value_c = (FlintType) 0;

      if (flags & UFLAG_SITES)
	{  /*  unit has sites  */
	FOR_ALL_SITES_AND_LINKS( unit_ptr, site_ptr, link_ptr )
          link_ptr->value_a = link_ptr->value_b = link_ptr->value_c = (FlintType) 0;
        }
      else
	{  /*  unit has no sites   */
	if (flags & UFLAG_DLINKS)
	  {  /*  unit has direct links	 */
	  FOR_ALL_LINKS( unit_ptr, link_ptr )
            link_ptr->value_a = link_ptr->value_b = link_ptr->value_c = (FlintType) 0;
          }
        }
      }
    }

  return( KRERR_NO_ERROR );
}



/*  quickprop backward error propagation
    (topological) for quickprop with SIGMOID_PRIME_OFFSET
    batch-modus: without adaption of links and bias
*/
static float  propagateNetBackwardQuickprop( pattern_no, delta_max )
int pattern_no;    /*  number of actual pattern  */
float  delta_max;  /*  delta max  */
{
  register struct Link *link_ptr;
  register struct Site *site_ptr;
  register struct Unit *unit_ptr;
  register Patterns out_pat;
  register float error,                    /*  error  */
                 sum_error,                /*  sum of the error  */
                 devit;                    /*  deviation  */
  TopoPtrArray topo_ptr;


  sum_error = 0.0;  /*  reset network error  */

  /*  calculate address of the output pattern (with number pattern_no + 1)  */
  out_pat = out_patterns + (pattern_no + 1) * NoOfOutputUnits;

  /*  add 3 to no_of_topo_units because the topologic array contains
      4 NULL pointers  */
  topo_ptr = topo_ptr_array + (no_of_topo_units + 3);

  /*  calculate output units only  */
  while ((unit_ptr = *--topo_ptr) != NULL)
    {
    devit = *(--out_pat) - unit_ptr->Out.output;
       /*= o * (1.0 - o) in [0.0,0.25],*/
       /*for asymmetric logistic function*/

    if ( (float) fabs( devit ) <= delta_max )  continue;

    sum_error += devit * devit;  /*  sum up the error of the network  */

    /*	calc. error for output units	 */
    error = devit * ((unit_ptr->act_deriv_func) ( unit_ptr ) + SIGMOID_PRIME_OFFSET);

    unit_ptr->value_c += -error /* * 1 */;   /*  calculate the bias slopes  */
                                            /*  learn bias like a weight  */
    if (UNIT_HAS_DIRECT_INPUTS( unit_ptr ))
      {  /*  the unit has direct links  */
      FOR_ALL_LINKS( unit_ptr, link_ptr )
        {  /*	calculate the slopes  */
        link_ptr->value_c += - error * link_ptr->to->Out.output;
        link_ptr->to->Aux.flint_no += link_ptr->weight * error;
        }
      }
    else
      {  /*  the unit has sites  */
      FOR_ALL_SITES_AND_LINKS( unit_ptr, site_ptr, link_ptr )
	{  /*  calculate the value_cs  */
	link_ptr->value_c += - error * link_ptr->to->Out.output;
	link_ptr->to->Aux.flint_no += link_ptr->weight * error;
	}
      }
    }


  /*  calculate hidden units only  */
  while ((unit_ptr = *--topo_ptr) != NULL)
    {
    error = ((unit_ptr->act_deriv_func) (unit_ptr) + SIGMOID_PRIME_OFFSET) * unit_ptr->Aux.flint_no;

    unit_ptr->value_c += - error /* * 1 */;   /*  calculate the bias slopes  */
                                            /*  learn bias like a weight  */
    if (UNIT_HAS_DIRECT_INPUTS( unit_ptr ))
      {  /*  the unit has direct links  */
      FOR_ALL_LINKS( unit_ptr, link_ptr )
	{  /*  calculate the slopes  */
	if (link_ptr->to->flags & UFLAG_TTYP_HIDD)
	  /*  this link points to a hidden unit: sum up the error's from previos units  */
	  link_ptr->to->Aux.flint_no += link_ptr->weight * error;

	link_ptr->value_c += - error * link_ptr->to->Out.output;
	}
      }
    else
      {  /*  the unit has sites  */
      FOR_ALL_SITES_AND_LINKS( unit_ptr, site_ptr, link_ptr )
	{  /*  calculate the slopes  */
	if (link_ptr->to->flags & UFLAG_TTYP_HIDD)
	  /*  this link points to a hidden unit: sum up the error's from previos units  */
	  link_ptr->to->Aux.flint_no += link_ptr->weight * error;

	link_ptr->value_c += - error * link_ptr->to->Out.output;
	}
      }
    }

 sum_error *= 0.5;
 return( sum_error );  /*  return the error of the network  */
}



static void MODI_quickprop( learn_parameter, /*modificate all weights and bias*/
                            max_factor,      /*of the net after every epoch*/
                            decay )
float learn_parameter,                    /*learning parameter*/
      max_factor,                         /*maximal grow factor of weights*/
      decay;                              /*decay factor*/

{
  double deltaw;			   /*actual weight (bias) change*/
  float shfac;				   /*shrink factor*/
  register struct Link *link_ptr;
  register struct Site *site_ptr;
  register struct Unit *unit_ptr;
  TopoPtrArray topo_ptr;
  bool hidden_units;


  /*  maximal grow factor of weights is max_factor  */
  shfac = max_factor / (1.0 + max_factor);

  topo_ptr = topo_ptr_array + (NoOfInputUnits + 1);
  hidden_units = TRUE;  

  /*  calculate hidden and output units only  */
  do
    {
    if ((unit_ptr = *++topo_ptr) == NULL)
      {
      if (!hidden_units)  break;  /*  end of topologic pointer array reached  */
      unit_ptr = *++topo_ptr;  /*  skip NULL pointer  */
      hidden_units = FALSE;
      }

    deltaw = 0.0;  /*  adjust bias like a weight  */
    if (unit_ptr->value_a > 0.0)
      {  /*  previous step was positive  */
      if (unit_ptr->value_c < 0.0)
	/*  same direction,i.e. slope, value_b have same sign  */
	deltaw += learn_parameter * (- unit_ptr->value_c);

      if (unit_ptr->value_c <= shfac * unit_ptr->value_b)
	/*  maximal positive step  */
	deltaw += max_factor * unit_ptr->value_a;
      else
	/*  littler positive step squared approximation  */
	deltaw += unit_ptr->value_c / (unit_ptr->value_b - unit_ptr->value_c)
		  * unit_ptr->value_a;
      }
    else
      if (unit_ptr->value_a < 0.0)
	{  /*  previous step was negative  */
	if (unit_ptr->value_c > 0.0)
	   /*  same direction,i.e. slope, prevslope have same sign  */
	  deltaw += learn_parameter * (- unit_ptr->value_c);

	if (unit_ptr->value_c >= shfac * unit_ptr->value_b)
	  /*  maximal negative step  */
	  deltaw += max_factor * unit_ptr->value_a;
	else
	  /* littler negative step squared approximation */
	  deltaw += unit_ptr->value_c / (unit_ptr->value_b - unit_ptr->value_c)
		    * unit_ptr->value_a;
	}
      else
	/*  previous step was 0.0  */
	/*  start of learning process with BP  */
	deltaw += learn_parameter * (- unit_ptr->value_c);

    unit_ptr->bias += deltaw;		    /*new bias*/
    unit_ptr->value_a = deltaw;	    /*bias change*/
    unit_ptr->value_b = unit_ptr->value_c;  /*previous slope*/
    unit_ptr->value_c = decay * unit_ptr->bias; /*  set new slope  */

    /*adjust links*/
    if (UNIT_HAS_DIRECT_INPUTS( unit_ptr ))
      {  /*  the unit has direct links	*/
      FOR_ALL_LINKS( unit_ptr, link_ptr )
	{
	deltaw = 0.0;

	if (link_ptr->value_a > 0.0)  /*  previous step was positive*/
	  {
	  if (link_ptr->value_c < 0.0)
	    /*	same direction,i.e. slope, prevslope have same sign  */
	    deltaw += learn_parameter * (- link_ptr->value_c);

	  if (link_ptr->value_c <= shfac * link_ptr->value_b)
	    /*	maximal positive step  */
	    deltaw += max_factor * link_ptr->value_a;
	  else
	    deltaw += link_ptr->value_c / (link_ptr->value_b - link_ptr->value_c)
		    * link_ptr->value_a;
	  }
	else
	  if (link_ptr->value_a < 0.0)  /*	previous step was negative*/
	    {
	    if (link_ptr->value_c > 0.0)
	      /*  same direction,i.e. slope, prevslope have same sign */
	      deltaw += learn_parameter * (- link_ptr->value_c);

	    if (link_ptr->value_c >= shfac * link_ptr->value_b)
	      /*  maximal negative step  */
	      deltaw += max_factor * link_ptr->value_a;
	    else
	      deltaw += link_ptr->value_c / (link_ptr->value_b - link_ptr->value_c)
			* link_ptr->value_a;
	    }
	  else	/*  previous step was 0.0  */
	    /*	start of learning process with BP  */
	    deltaw += learn_parameter * (- link_ptr->value_c);

	link_ptr->weight += deltaw;	      /*new weight*/
	link_ptr->value_a = deltaw;       /*weight change*/
	link_ptr->value_b = link_ptr->value_c;	/*previous slope*/
	/*  set new slope  */
	link_ptr->value_c = decay * link_ptr->weight;
	}  /*  for links  */
      }
    else
      {  /*  the unit has sites  */
      FOR_ALL_SITES_AND_LINKS( unit_ptr, site_ptr, link_ptr )
	{
	deltaw = 0.0;
	if (link_ptr->value_a > 0.0)
	  {  /*  previous step was positive  */
	  if (link_ptr->value_c < 0.0)
	    /*	same direction,i.e. slope, prevslope have same sign  */
	    deltaw += learn_parameter * (- link_ptr->value_c);

	  if (link_ptr->value_c <= shfac * link_ptr->value_b)
	    /*	maximal positive step  */
	    deltaw += max_factor * link_ptr->value_a;
	  else
	    /*	littler positive step squared approximation  */
	    deltaw += link_ptr->value_c / (link_ptr->value_b - link_ptr->value_c)
		      * link_ptr->value_a;
	  }
	else
	  if (link_ptr->value_a < 0.0)
	    {  /*  previous step was negative  */
	    if (link_ptr->value_c > 0.0)
	      /*  same direction,i.e. slope, prevslope have same sign  */
	      deltaw += learn_parameter * (- link_ptr->value_c);

	    if (link_ptr->value_c >= shfac * link_ptr->value_b)
	      /*  maximal negative step  */
	      deltaw += max_factor * link_ptr->value_a;
	    else
	      deltaw += link_ptr->value_c / (link_ptr->value_b - link_ptr->value_c)
			* link_ptr->value_a;
	    }
	  else	/*  previous step was 0.0  */
	    /*	start of learning process with BP  */
	    deltaw += learn_parameter * (- link_ptr->value_c);

	link_ptr->weight += deltaw;	     /*new weight*/
	link_ptr->value_a = deltaw;      /*weight change*/
	link_ptr->value_b = link_ptr->value_c;	/*previous slope*/
	/*set new slope*/
	link_ptr->value_c = decay * link_ptr->weight;
	}
      }
    }  /*  for units  */
  while( TRUE );

}


/*  Quickprop learning function
    Input Parameters:   1 : learning parameter
                        2 : max factor (of the net after every epoch)
                        3 : decay
                        4 : delta max

    Output Parameters:  1 : error of the network (sum of all cycles)
*/
krui_err LEARN_quickprop( start_pattern, end_pattern,
			 parameterInArray, NoOfInParams,
			 parameterOutArray, NoOfOutParams )
int  start_pattern, end_pattern;
float  parameterInArray[], * *parameterOutArray;
int  NoOfInParams, *NoOfOutParams;
{
  static float OutParameter[1];           /*OutParameter[0] stores the*/
                                          /*learning error*/
  int  i, pattern_no, ret_code;


  if (NoOfUnits == 0)
    return( KRERR_NO_UNITS );  /*  No Units defined  */

  if (NoOfInParams < 1)    /* #########   have to be changed  (must be 3)  ######### */
    return( KRERR_PARAMETERS );            /*  not enough input parameters  */

  *NoOfOutParams = 1;                      /*  one return value is available (the learning error) */

  *parameterOutArray = OutParameter;       /*  set output parameter reference  */
  ret_code = KRERR_NO_ERROR;               /*  reset return code  */

  if (NetModified || (TopoSortID != TOPOLOGICAL_FF))
    {  /*  Net has been modified or topologic array isn't initialized */
    /*  check the topology of the network  */
    ret_code = kr_topoCheck();
    if (ret_code < KRERR_NO_ERROR)  return( ret_code );  /*  an error has occured  */
    if (ret_code < 2)  return( KRERR_NET_DEPTH );  /*  the network has less then 2 layers  */

    /*	count the no. of I/O units and check the patterns  */
    ret_code = kr_IOCheck();
    if (ret_code < KRERR_NO_ERROR)  return( ret_code );

    /*	sort units by topology and by topologic type  */
    ret_code = kr_topoSort( TOPOLOGICAL_FF );
    if ((ret_code != KRERR_NO_ERROR) && (ret_code != KRERR_DEAD_UNITS))
      return( ret_code );

    NetModified = FALSE;
    }

  if (NetInitialize || LearnFuncHasChanged)
    {  /*  Net has been modified or initialized, initialize backprop now  */
    ret_code = initializeQuickprop();
    if (ret_code != KRERR_NO_ERROR)  return( ret_code );
    }


  NET_ERROR( OutParameter ) = 0.0;   /*  reset network error value  */
  for (i = start_pattern; i <= end_pattern; i++)
    {
    if (NoOfShuffledPatterns == NoOfPatternPairs)
      pattern_no = pattern_numbers[i];  /*  shuffle patterns   */
    else
      pattern_no = i;  /*  do not schuffle patterns  */

    propagateNetForward( pattern_no );         /*  forward propagation  */

    /*  backward propagation and summation of gradient  */
    NET_ERROR(OutParameter) +=
      propagateNetBackwardQuickprop( pattern_no,
                                     LEARN_PARAM4( parameterInArray ) );
    }

  /*  modificate links and bias  */
  MODI_quickprop( LEARN_PARAM1( parameterInArray ),
                  LEARN_PARAM2( parameterInArray ),
                  LEARN_PARAM3( parameterInArray ) );

  return( ret_code );
}



/*#################################################

GROUP: Counterpropagation learning function

#################################################*/


/*  Counterpropagation initialisation
*/
static krui_err  initializeCPN()
{
  register struct Unit   *unit_ptr;


  NoOfLearnedPatterns = 0;

  /*  set unit's bias to zero  */
  FOR_ALL_UNITS( unit_ptr )
    if ( (unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE)
      /*  unit is in use  */
      unit_ptr->bias = (FlintType) 0;

  return( KRERR_NO_ERROR );
}


static void normalize_weight( winner_ptr, sum )
struct Unit *winner_ptr;
float       sum;
{
  register struct Site *site_ptr;
  register struct Link *link_ptr;
  register float amount;


  amount = 1.0 / sqrt( sum );

  if (winner_ptr->flags & UFLAG_SITES)
    /* the unit has sites */
    FOR_ALL_SITES_AND_LINKS( winner_ptr, site_ptr, link_ptr )
      link_ptr->weight = link_ptr->weight * amount;
  else 
    /* the unit has direct links */
    FOR_ALL_LINKS( winner_ptr, link_ptr )
      link_ptr->weight = link_ptr->weight * amount;
}


static void normalize_inputvector( sum )
float sum;
{
  register struct Unit *unit_ptr;
  register float amount;


  amount = 1.0 / sqrt( sum );

  FOR_ALL_UNITS( unit_ptr )
    if (((unit_ptr->flags & UFLAG_TTYP_PAT) == UFLAG_TTYP_IN) &&
        ((unit_ptr->flags & UFLAG_IN_USE) == UFLAG_IN_USE))
      /* this is a input unit */
      unit_ptr->Out.output = unit_ptr->Out.output * amount;
}  


/************************************************************************
 *                   function propagateNet_CPN                          *
 ************************************************************************
 *                                                                      *
 ************************************************************************/

static float propagateNet_CPN( pattern_no, alpha, beta, threshold )
int    pattern_no;
float  alpha;
float  beta;
float  threshold;
{
  register struct Link    *link_ptr;
  register struct Site    *site_ptr;
  register struct Unit    *unit_ptr;
  register struct Unit    *winner_ptr;
  register Patterns       in_pat, out_pat;
  float          maximum, sum_error, devit, learn_error, sum;
  float          unit_ptr_net;
  float          noOfPatterns_mul_NoHiddenUnits;
  register TopoPtrArray     topo_ptr;


  /***************************************************************/
  /*      calculate the activation and the output values         */
  /*      of the input units (Input Layer)                       */
  /***************************************************************/ 

  noOfPatterns_mul_NoHiddenUnits = (float) NoOfLearnedPatterns * (float) NoOfHiddenUnits;

  sum = 0.0;

  /*  calculate startaddress for input pattern array  */
  in_pat = in_patterns + pattern_no * NoOfInputUnits;

  topo_ptr = topo_ptr_array;

  /*  copy pattern into input unit's activation and
      calculate output of the input units
  */
  while ((unit_ptr = *++topo_ptr) != NULL)
    {  /*  topo_ptr points to the unit stuctures
       (sorted by: input-, hidden- and output-units, separated with NULL pointers)
       */
    sum += *in_pat * *in_pat;

    if (unit_ptr->out_func == OUT_IDENTITY)
      /*  identity output function: there is no need to call the output function  */
      unit_ptr->Out.output = unit_ptr->act = *in_pat++;
    else
      /*  no identity output function: calculate unit's output also  */
      unit_ptr->Out.output = (*unit_ptr->out_func) (unit_ptr->act = *in_pat++);
    }

   if (sum != 0.0)
     /* normalize the inputvector */
     normalize_inputvector( sum );


  /***************************************************************/
  /*              propagate Kohonen Layer                        */
  /***************************************************************/

  /***************************************************************/
  /*      calculate the activation and the output values         */
  /*      of the hidden units (Kohonen Layer)                    */
  /***************************************************************/

  winner_ptr = NULL;
  maximum = -1.0e30;  /* contains the maximum of the activations */

  /*  popagate hidden units  */
  while ((unit_ptr = *++topo_ptr) != NULL)
    {  /*  topo_ptr points to a (topological sorted) unit stucture */
    unit_ptr_net = 0.0;
    if (UNIT_HAS_DIRECT_INPUTS( unit_ptr ))
      {  /* the unit has direct links */
      FOR_ALL_LINKS( unit_ptr, link_ptr )
        unit_ptr_net += (link_ptr->weight * link_ptr->to->Out.output);
      }
    else
      { /*  the unit has sites	*/
      FOR_ALL_SITES_AND_LINKS( unit_ptr, site_ptr, link_ptr )
	unit_ptr_net += (link_ptr->weight * link_ptr->to->Out.output);
      }

    if (unit_ptr->bias >= noOfPatterns_mul_NoHiddenUnits)
      unit_ptr_net -= threshold;

    if (maximum < unit_ptr_net)
      {  /*  determine winner unit  */
      winner_ptr = unit_ptr;
      maximum = unit_ptr_net;
      }

    /*  reset output and activation of hidden units  */
    unit_ptr->Out.output = unit_ptr->act = (FlintType) 0;
    }

  /***************************************************************/
  /*             the competitive winner is chosen                */
  /***************************************************************/
  winner_ptr->Out.output = winner_ptr->act = (FlintType) 1;
  winner_ptr->bias++;


  /***************************************************************/
  /*                Training the Kohonen Layer                   */
  /*                                                             */
  /* Only the weights of links that go to the winning unit are   */
  /* adjusted, the others remain the same.                       */
  /* The incoming weights to the competitive unit are adapted as */
  /* follows:                                                    */
  /*                                                             */
  /* weight(new) = weight(old) + alpha * (output - weight(old))  */
  /*                                                             */
  /* where eta is the learning constant (0<eta<=1.0)             */
  /* and output is the output of the input unit                  */
  /*                                                             */
  /***************************************************************/


  sum = 0.0;
  if (winner_ptr->flags & UFLAG_DLINKS)
    {  /*  the winner unit has direct links  */
    FOR_ALL_LINKS( winner_ptr, link_ptr )
      {
      devit             = link_ptr->to->Out.output - link_ptr->weight;
      learn_error       = alpha * devit;
      link_ptr->weight += learn_error;        
      /* this is needed for the normalization of the weight_vector */
      sum              += link_ptr->weight * link_ptr->weight;
      }
    }
  else
    {  /*  the winner unit has sites  */
    FOR_ALL_SITES_AND_LINKS( winner_ptr, site_ptr, link_ptr )
      {
      devit		= link_ptr->to->Out.output - link_ptr->weight;
      learn_error	= alpha * devit;
      link_ptr->weight += learn_error;
      /* this is needed for the normalization of the weight_vector */
      sum	       += link_ptr->weight * link_ptr->weight;
      }
    }

  if (sum != 0.0)
     normalize_weight( winner_ptr, sum );

  /***************************************************************/
  /*              propagate Grossberg Layer                      */
  /***************************************************************/
  /*                    AND                                      */
  /***************************************************************/
  /*              Training the Grossberg Layer                   */
  /*                                                             */
  /* Adaptation of the Grossberg Layer weights is done by the    */
  /* Widrow-Hoff rule:                                           */
  /*                                                             */
  /* weight(new) = weight(old) + beta * (target output - output) */
  /*                                                             */
  /* for all weights connected with the winning unit of the      */
  /* Kohonen Layers                                              */
  /*                                                             */
  /***************************************************************/




  /* calculate address of the output pattern */
  out_pat = out_patterns + pattern_no * NoOfOutputUnits;

  sum_error = 0.0;

  /*  popagate output units  */
  while ((unit_ptr = *++topo_ptr) != NULL)
    {  /*  topo_ptr points to a (topological sorted) unit stucture */

    /***************************************************************/
    /*      calculate the activation and the output values         */
    /*      of the output units (Grossberg Layer)                  */
    /***************************************************************/
    /* the activation function is the identity function (weighted sum)
       and identity output function
    */
    unit_ptr->Out.output = unit_ptr->act = (*unit_ptr->act_func) (unit_ptr);

    devit       = *out_pat++ - unit_ptr->Out.output; /* calculate devitation */
    sum_error  += devit * devit;
    learn_error = beta * devit ;

    if (UNIT_HAS_DIRECT_INPUTS( unit_ptr ))
      {  /* the unit has direct links */
      FOR_ALL_LINKS( unit_ptr, link_ptr )
        if ( link_ptr->to == winner_ptr )
          {
           /* link to the winning unit of the Kohonen Layer */
           link_ptr->weight += learn_error;
           break;
	  }
      }
    else
      {  /* the unit has sites */
      FOR_ALL_SITES_AND_LINKS( unit_ptr, site_ptr, link_ptr )
	if ( link_ptr->to == winner_ptr )
	  {
	  /* link to the winning unit of the Kohonen Layer */
	  link_ptr->weight  += learn_error;
	  break;
          }
      }

    }

  sum_error *= 0.5; 
  return( sum_error );
}



krui_err  LEARN_CPN( start_pattern, end_pattern, parameterInArray, NoOfInParams, parameterOutArray, NoOfOutParams )
int     start_pattern, end_pattern;
float   parameterInArray[],
        * *parameterOutArray;
int  NoOfInParams,
     *NoOfOutParams;
{
  static  float  OutParameter[1];   /*  OutParameter[0] stores the learning error  */
  int     ret_code, i, pattern_no;


  if (NoOfUnits == 0)
    return( KRERR_NO_UNITS );    /*  No Units defined    */
  if (NoOfInParams < 1)          /*  have to be changed (must be 3) */
    return( KRERR_PARAMETERS );  /*  Not enough input parameters  */

  *NoOfOutParams = 1;                   /*  one return value is available (the learning error) */
  *parameterOutArray = OutParameter;    /*  set output parameter reference  */
  ret_code = KRERR_NO_ERROR;            /*  clear return code  */


  if (NetModified || (TopoSortID != TOPOLOGIC_TYPE))
    {  /*  Net has been modified or topologic array isn't initialized */
    /*  check the topology of the network  */
    ret_code = kr_topoCheck();
    if (ret_code < KRERR_NO_ERROR)  return( ret_code );  /*  an error has occured  */
    if (ret_code != 3)	return( KRERR_NET_DEPTH );  /*	the network has less then 2 layers  */

    /*	count the no. of I/O units and check the patterns  */
    ret_code = kr_IOCheck();
    if (ret_code < KRERR_NO_ERROR)  return( ret_code );

    /*	sort units by topology and by topologic type  */
    ret_code = kr_topoSort( TOPOLOGIC_TYPE );
    if ((ret_code != KRERR_NO_ERROR) && (ret_code != KRERR_DEAD_UNITS))
      return( ret_code );

    NetModified = FALSE;
    }

  if (NetInitialize || LearnFuncHasChanged)
    {  /*  Net has been modified or initialized, initialize backprop now  */
    ret_code = initializeCPN();
    if (ret_code != KRERR_NO_ERROR)  return( ret_code );
    }


  NET_ERROR( OutParameter ) = 0.0;   /*  reset network error value  */
  for (i = start_pattern; i <= end_pattern; i++)
    {
    if (NoOfShuffledPatterns == NoOfPatternPairs)
      pattern_no = pattern_numbers[i];  /*  shuffle patterns   */
    else
      pattern_no = i;  /*  do not schuffle patterns  */

    NoOfLearnedPatterns++;
    NET_ERROR( OutParameter ) +=
      propagateNet_CPN( pattern_no,
                        LEARN_PARAM1( parameterInArray ),
                        LEARN_PARAM2( parameterInArray ),
                        LEARN_PARAM3( parameterInArray ) );
    }

  return( ret_code );
}




/* *******************************************************

  Back-Percolation Learning Function
  by Artemis Hatzigeorgiou
  Algorithm by Mark Jurik

*/

/*  topological forward propagation
*/
static	void  propagateNetForward_perc( pattern_no )
int     pattern_no;
{
  register struct Unit   *unit_ptr;
  register Patterns  in_pat;
  register TopoPtrArray     topo_ptr;


  /*  calculate startaddress for input pattern array  */
  in_pat = in_patterns + pattern_no * NoOfInputUnits;

  topo_ptr = topo_ptr_array;

  /*  copy pattern into input unit's activation and
      calculate output of the input units
  */
  while ((unit_ptr = *++topo_ptr) != NULL)
    {  /*  topo_ptr points to a (topological sorted) unit stucture (input units first)  */
    if (unit_ptr->out_func == OUT_IDENTITY)
      /*  identity output function: there is no need to call the output function  */
      unit_ptr->Out.output = unit_ptr->act = *in_pat++;
    else
      /*  no identity output function: calculate unit's output also  */
      unit_ptr->Out.output = (*unit_ptr->out_func) (unit_ptr->act = *in_pat++);
    }

  /*  popagate hidden units  */
  while ((unit_ptr = *++topo_ptr) != NULL)
    {  /*  topo_ptr points to a (topological sorted) unit stucture */
    /*  clear values  */
    unit_ptr->Aux.flint_no = 0.0;
    unit_ptr->value_a = 0.0;
    unit_ptr->value_b = 0.000001;

    /*  calculate the activation value of the unit: call the activation function if needed  */
    unit_ptr->act = (*unit_ptr->act_func) (unit_ptr);

    if (unit_ptr->out_func == OUT_IDENTITY)
      /*  identity output function: there is no need to call the output function  */
      unit_ptr->Out.output = unit_ptr->act;
    else
      /*  no identity output function: calculate unit's output also  */
      unit_ptr->Out.output = (*unit_ptr->out_func) (unit_ptr->act);
    }

  /*  popagate output units  */
  while ((unit_ptr = *++topo_ptr) != NULL)
    {  /*  topo_ptr points to a (topological sorted) unit stucture */
    /*  clear values  */
    unit_ptr->Aux.flint_no = 0.0;
    unit_ptr->value_a = 0.0;
    unit_ptr->value_b = 0.000001;

    /*  calculate the activation value of the unit: call the activation function if needed  */
    unit_ptr->act = (*unit_ptr->act_func) (unit_ptr);

    if (unit_ptr->out_func == OUT_IDENTITY)
      /*  identity output function: there is no need to call the output function  */
      unit_ptr->Out.output = unit_ptr->act;
    else
      /*  no identity output function: calculate unit's output also  */
      unit_ptr->Out.output = (*unit_ptr->out_func) (unit_ptr->act);
    }
}


/*  Backward error propagation (topological)
*/
static	float propagateNetBackward_perc( pattern_no, learn_parameter, delta_max ,perc_error)
int     pattern_no;
float   learn_parameter, delta_max,  *perc_error;
{
  register struct Link   *link_ptr;
  register struct Site   *site_ptr;
  register struct Unit   *unit_ptr;
  register Patterns      out_pat;
  register float         error,  sum_error,  eta,  devit, learn_error;
  register TopoPtrArray     topo_ptr;
  register  float     weight, norm, delta_sig_normaliser, message_weight; 
  register  float     act_err, normalised_error, scaled_error, delta_weight_normaliser;
  register  float     der = 0.0;

  sum_error = 0.0;    /*  reset network error  */
  eta = learn_parameter;   /*  store learn_parameter in CPU register  */
  /*  calculate address of the output pattern (with number pattern_no + 1)  */
  out_pat = out_patterns + (pattern_no + 1) * NoOfOutputUnits;

  /*  add 3 to no_of_topo_units because the topologic array contains
      4 NULL pointers  */
  topo_ptr = topo_ptr_array + (no_of_topo_units + 3);

  /*  calculate output units only  */
  while ((unit_ptr = *--topo_ptr) != NULL)
    {
    devit = *(--out_pat) - unit_ptr->Out.output;  /*  calc. devitation	*/

    if ( fabs( devit ) > delta_max )
      {  /*  calc. error for output units     */
      *perc_error += fabs(devit);
      error = -2.0 * devit * (unit_ptr->act_deriv_func) ( unit_ptr );
      act_err = devit*eta; 
      sum_error += devit * devit;  /*  sum up the error of the network	*/
      }
    else
      {  /*  set error of output units to zero	*/
      error = 0.0;
      act_err = devit*eta; 
      }

    /*	calc. the error for adjusting weights and bias of the predecessor units  */
  

        norm = 0.0;
        delta_sig_normaliser = 0.000001;     
        FOR_ALL_LINKS( unit_ptr, link_ptr )
	  {  /*  adjust link weights and calc. sum of errors of the predecessor units  */
          link_ptr->to->Aux.flint_no += link_ptr->weight * error;
          if ((link_ptr->to->flags & UFLAG_TTYP_PAT) != UFLAG_TTYP_IN)
            norm += fabs( link_ptr->weight );
          delta_sig_normaliser += SQR(link_ptr->to->Out.output);
	  }
        delta_weight_normaliser =   delta_sig_normaliser + 1;
        norm += delta_sig_normaliser;
        normalised_error = act_err/norm; 
        scaled_error = act_err/delta_weight_normaliser;
        FOR_ALL_LINKS( unit_ptr, link_ptr )    
          {         
          message_weight = (error * link_ptr->weight) * (error * link_ptr->weight);
          link_ptr->to->value_a += link_ptr->weight * normalised_error * message_weight;
          link_ptr->to->value_b += message_weight;
          link_ptr->weight += link_ptr->to->Out.output * scaled_error;
	  }


    /*	adjust bias value  */
    unit_ptr->bias += scaled_error;
    }

  /*  calculate hidden units only  */
  while ((unit_ptr = *--topo_ptr) != NULL)
    {
    der =  (unit_ptr->act_deriv_func) ( unit_ptr );
    error = der * unit_ptr->Aux.flint_no;
    act_err = (unit_ptr->value_a / unit_ptr->value_b) * der ;  
    
    /*	calc. the error for adjusting weights and bias of the predecessor units  */
 

      norm = 0.0;
      delta_sig_normaliser = 0.000001;    
      FOR_ALL_LINKS( unit_ptr, link_ptr )
	{ 
        link_ptr->to->Aux.flint_no += link_ptr->weight * error;
	if (link_ptr->to->flags & UFLAG_TTYP_HIDD)
            norm += fabs( link_ptr->weight );

        delta_sig_normaliser += SQR(link_ptr->to->Out.output);
        }
      delta_weight_normaliser =   delta_sig_normaliser + 1;
      norm += delta_sig_normaliser;
      normalised_error = act_err/norm; 
      scaled_error = act_err/delta_weight_normaliser;
      FOR_ALL_LINKS( unit_ptr, link_ptr ) 
        {
	message_weight = (error * link_ptr->weight) * (error * link_ptr->weight);
        link_ptr->to->value_a += link_ptr->weight * normalised_error * message_weight;
        link_ptr->to->value_b += message_weight;
        link_ptr->weight += link_ptr->to->Out.output * scaled_error;
        }


    /*	adjust bias value  */
    unit_ptr->bias += scaled_error;
   }


  sum_error *= 0.5;
  return( sum_error );  /*  return the error of the network */
}


/*  BackPercolation Learning Function
    Input Parameters:   1 : learning parameter
                        2 : delta max

    Output Parameters:  1 : error of the network (sum of all cycles)

*/
krui_err  LEARN_perc( start_pattern, end_pattern, parameterInArray, NoOfInParams, parameterOutArray, NoOfOutParams )
int     start_pattern, end_pattern;
float   parameterInArray[],
        * *parameterOutArray;
int  NoOfInParams,
     *NoOfOutParams;
{
  static  float  OutParameter[1];   /*	OutParameter[0] stores the learning error  */
  int	  i, ret_code, pattern_no;
  float   p_error,l_error;
  register struct Unit   *unit_ptr;

  if (NoOfUnits == 0)  return( KRERR_NO_UNITS );  /*  No Units defined	*/
  if (NoOfInParams < 1)          /*  ####  have to be changed (must be 2)  #### */
    return( KRERR_PARAMETERS );  /*  Not enough input parameters  */

  *NoOfOutParams = 1;   /*  One return value is available (the learning error)  */
  *parameterOutArray = OutParameter;   /*  set the output parameter reference  */ 
  ret_code = KRERR_NO_ERROR;  /*  reset return code  */

  if (NetModified || (TopoSortID != TOPOLOGICAL_FF))
    {  /*  Net has been modified or topologic array isn't initialized */
      /*  check the topology of the network  */
        FOR_ALL_UNITS(unit_ptr)
        {
           if(UNIT_HAS_SITES( unit_ptr ))
           {    return(KRERR_PERC_SITES);}
         }
    
    ret_code = kr_topoCheck();
    if (ret_code < KRERR_NO_ERROR)  return( ret_code );  /*  an error has occured  */
    if (ret_code < 2)  return( KRERR_NET_DEPTH );  /*  the network has less then 2 layers  */

    /*	count the no. of I/O units and check the patterns  */
    ret_code = kr_IOCheck();
    if (ret_code < KRERR_NO_ERROR)  return( ret_code );

    /*	sort units by topology and by topologic type  */
    ret_code = kr_topoSort( TOPOLOGICAL_FF );
    if ((ret_code != KRERR_NO_ERROR) && (ret_code != KRERR_DEAD_UNITS))
      return( ret_code );

    NetModified = FALSE;
    }

  if (NetInitialize || LearnFuncHasChanged)
    {  /*  Net has been modified or initialized, initialize backprop now  */
    ret_code = initializeBackprop();
    if (ret_code != KRERR_NO_ERROR)  return( ret_code );
    parameterInArray[1] = 1.0;
    }

  NET_ERROR( OutParameter ) = 0.0;   /*  reset network error value  */
  p_error = 0.0;
  for (i = start_pattern; i <= end_pattern; i++)
    {
    if (NoOfShuffledPatterns == NoOfPatternPairs)
      pattern_no = pattern_numbers[i];  /*  shuffle patterns   */
    else
      pattern_no = i;  /*  do not schuffle patterns  */

    propagateNetForward_perc( pattern_no );	/*  Forward propagation */

    /*  Backward propagation  */
    /*  1st parameter is the learning parameter
        2nd parameter is the max. devitation between output pattern and the output of the output unit
        (delta max)
    */

      NET_ERROR( OutParameter ) +=
	propagateNetBackward_perc( pattern_no,
			      LEARN_PARAM1( parameterInArray ),
                              LEARN_PARAM5( parameterInArray ), &p_error ); 
    }

     p_error = p_error /( NoOfPatternPairs * NoOfOutputUnits ) ;
    l_error = exp( ( parameterInArray[1] - p_error ) / (parameterInArray[1] + p_error ) ); 
    parameterInArray[1] = p_error;

    if( l_error <= 0.5)    l_error = 0.5;
    else  if( l_error >= 1.05)   l_error = 1.05; 
 
  
    if( ( l_error = parameterInArray[0] * l_error ) < 0.8)
          parameterInArray[0] = l_error;
   return( ret_code );
}

