/*     %W% %G%  */

/**********************************************************************
FILE   : %M%
PURPOSE: SNNS Batchlearning Program
NOTES  : 
AUTHOR : Niels Mache
DATE   : 26.6.91
VERSION : %I%  %G%

 Copyright (c) 1990,1991,1992 by Niels Mache and the SNNS Group

**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <ctype.h>
#include <limits.h>
#include <time.h>
#include <string.h>
#include <fcntl.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/utsname.h>


/*  SNNS-Kernel constants and data type definitions  */
#include "glob_typ.h"
/*  SNNS-Kernel User-Interface Function Prototypes   */
#include "kr_ui.h"


#ifdef __STDC__
extern int unlink(char const *);
#endif

/*  save temporary networkfile every 1800 seconds of cpu time */
#define SAVE_PERIOD 1800
/*  display not more than 100 cycles  (reduce size of the logfile)  */
#define NO_OF_CYCLES_DISPLAYED  100

/*  longest path name  */
#ifndef PATH_MAX
#define PATH_MAX 512
#endif

#define ABNORMAL  TRUE
#define NORMAL	  FALSE



/*  globals  */
FILE  *file_out = NULL;
char  *tmp_file1, *tmp_file2, *netname;
char  log_file[PATH_MAX + 1];
long  starttime, HZ_value;
int   no_of_sites, no_of_links, no_of_units, tmp_switch, cycle,
      no_of_patterns, no_of_return_values, tmp_switch;
float *return_values;
struct tms  tp;





/*  skip comments  */
static	bool	skipComments( stream )
register  FILE	*stream;
{
  register int   c;


  while (TRUE)
    {
    do
      c = getc( stream );
    while ( isspace( c ) );

    if (c != '#')  break;

    do
      c = getc( stream );
    while (c != EOF && c != '\n');
  }

  if (c == EOF)  return( FALSE );

  ungetc( c, stream );
  return( TRUE );
}


/*  Check for errors  */
void    errChk( errNo )
int     errNo;
{
  if (errNo != 0)
    {
    fprintf( stderr, "Batchlearning: %s\n", krui_error( errNo ));
    if (file_out != NULL)
      fprintf( file_out, "Batchlearning: %s\n", krui_error( errNo ));
    exit( 1 );
    }
}

void  writeLearnFuncValues()
{
  int  j;


  fprintf( file_out, "\nCycle: %d\nLearning function value(s): ", cycle );
  for (j = 0; j < no_of_return_values; j++ )
    fprintf( file_out, "[%d]: %g  ", j + 1, return_values[ j ] );
  fprintf( file_out, "\n" );
}


void  writeInfo( stream )
FILE  *stream;
{
  long	clock;
  struct utsname name;

  if (stream != NULL)
    {
    clock = 1;
    (void) time( &clock);

    fprintf( stream, "\n%s Batchlearning terminated at %s",
	     krui_getVersion(), ctime( &clock) );
    if (uname( &name ) == 0)
      fprintf( stream, "System: %s  Node: %s  Machine: %s\n",
	       name.sysname, name.nodename, name.machine );
  }
}



/*  write Statistics  */
void  writeStatistics( status )
bool  status;
{
  long	clock;
  float cpu_time;


  if (file_out != NULL)
    {
    clock = 1;
    (void) time( &clock);
    (void) times( &tp );
    cpu_time = (float) tp.tms_utime / (float) HZ_value;

    writeInfo( file_out );
    fprintf( file_out, "********************************************************\n" );
    fprintf( file_out, "\n----   STATISTICS  ----\n" );
    fprintf( file_out, "No. of learned cycles: %d\n", cycle );
    if (status)  writeLearnFuncValues();
    fprintf( file_out, "No. of units updated : %d\n", no_of_units * cycle * no_of_patterns);
    fprintf( file_out, "No. of sites updated : %d\n", no_of_sites * cycle * no_of_patterns);
    fprintf( file_out, "No. of links updated : %d\n", no_of_links * cycle * no_of_patterns);
    fprintf( file_out, "CPU Time used: %.2f seconds\n", cpu_time );
    fprintf( file_out, "User time: %d seconds\n", clock - starttime );
    fprintf( file_out, "No. of connection updates per second (CUPS): %e",
	   ((double) no_of_links * (double) cycle * (double) no_of_patterns) /
	   ((double) cpu_time) );

  }
}

static void  writeCaughtMsg()
{
  fprintf( stdout, "\nSignal caught, %s Batchlearning terminated.\n",
	   krui_getVersion() );
  writeInfo( stdout );

  if (file_out != NULL)
    {
    fprintf( file_out, "\n#####  Signal caught, %s Batchlearning terminated.\n",
	     krui_getVersion() );
    writeInfo( file_out );
  }
}


static void  emergency_exit()
{
  writeCaughtMsg();

  fprintf( stdout, "\nFatal Error in SNNS-Kernel. Networkfile is lost.\n" );

  if (file_out != NULL)
    {  fprintf( stdout, "\n#####  Fatal Error in SNNS-Kernel. Networkfile is lost.\n" );
       writeStatistics( ABNORMAL );
       fprintf( stdout, "Logfile '%s' written.\n", log_file );
       fclose( file_out );
    }

  exit( -3 );
}




/*  handle signals  */
static void  exit_batchlearn()
{
  char	*tmp_file;


  writeCaughtMsg();

  if (krui_getNoOfUnits() > 0)
    {  /*  save network  */
    if (tmp_switch & 1)
      tmp_file = tmp_file1;
    else
      tmp_file = tmp_file2;

    (void) krui_saveNet( tmp_file, netname );
    fprintf( stdout, "Networkfile '%s' saved.\n", tmp_file );
    if (file_out != NULL)
      fprintf( file_out, "Networkfile '%s' saved.\n", tmp_file );
    }

  if (file_out != NULL)
    {  writeStatistics( ABNORMAL );
       fprintf( stdout, "Logfile '%s' written.\n", log_file );
       fclose( file_out );
    }

  exit( -2 );
}


/*  catch these signals  */
void  catchSignals()
{
  signal( SIGHUP,  exit_batchlearn);  signal( SIGINT,  exit_batchlearn);
  signal( SIGQUIT, exit_batchlearn);  signal( SIGALRM, exit_batchlearn);
  signal( SIGTERM, exit_batchlearn);

#ifdef SIGPWR
  signal( SIGPWR,  exit_batchlearn);
#endif

  signal( SIGILL,  emergency_exit );
  signal( SIGTRAP, emergency_exit );  signal( SIGIOT,  emergency_exit );
  signal( SIGEMT,  emergency_exit );  signal( SIGFPE,  emergency_exit );
  signal( SIGBUS,  emergency_exit );  signal( SIGSEGV, emergency_exit );
}


/*  read the shuffleing ON/OFF switch  */
bool  getShuffleingSwitch( file_in )
FILE  *file_in;
{
  int  i, ret;
  char	do_shuffleing[80];


  skipComments( file_in );
  ret = fscanf( file_in, "%s", do_shuffleing );
  if (ret != 1)
    {
    fprintf( stderr, "Missing shuffleing switch (Expecting ON or OFF).\n" );
    exit( 1 );
    }
  for (i = 0; i < strlen( do_shuffleing ); i++)
    do_shuffleing[i] = toupper( do_shuffleing[i] );

  if (strcmp( do_shuffleing, "ON" ) == 0)
    {  fprintf( file_out, "Shuffleing of patterns enabled.\n" );
       return( TRUE );
    }
  else if (strcmp( do_shuffleing, "OFF" ) == 0)
	 {  fprintf( file_out, "Shuffleing of patterns disabled.\n" );
	    return( FALSE );
	 }
       else
	 {  fprintf( stderr, "Wrong shuffleing switch (Expecting ON or OFF).\n" );
	    exit( 1 );
	 }

  return( FALSE );
}



int main( argc, argv )
int  argc;
char  *argv[];
{
  int	ret, i, no_of_cycles, step;
  bool	shuffle;
  char	config_file[PATH_MAX + 1], pat_file[PATH_MAX + 1],
	net_file[PATH_MAX + 1];
  float learn_parameters[NO_OF_LEARN_PARAMS], best_net_error;
  long	clock;
  time_t lasttime;
  FILE	*file_in;


  fprintf( stdout, "\n%s\n", krui_getVersion() );
  fprintf( stdout, "-----  Batchlearning  -----\n\n" );

  switch (argc)
    {
    case 1:
      strcpy( config_file, "snnsbat.cfg" );
      strcpy( log_file, "snnsbat.log" );
      break;
    case 2:
      strcpy( config_file, argv[1] );
      strcpy( log_file, "snnsbat.log" );
      break;
    case 3:
      strcpy( config_file, argv[1] );
      strcpy( log_file, argv[2] );
      break;
    default:
      fprintf( stdout, "Usage: batch [config_file] [log_file]\n" );
      exit( 1 );
    }


  tmp_switch = 0;
  cycle = 0;


  /*  open Log-file  */
  if ((file_out = fopen( log_file, "w" )) == NULL)
    {
    fprintf( stderr, "Can't open LOG file\n" );
    exit( 1 );
    }

  fprintf( file_out, "%s Batchlearning Program\n", krui_getVersion() );

  /*  get HZ value  */
  if ((getenv( "HZ" ) == NULL) ||
      (sscanf( getenv( "HZ" ), "%d", &HZ_value ) != 1))
    {
#ifndef HZ
    fprintf( stderr, "****  WARNING: no HZ defined!\n" );
    HZ_value = 100;
#else
    HZ_value = HZ;
#endif
  }

  /*  open configuration file  */
  if ((file_in = fopen( config_file, "r" )) == NULL)
    {
    fprintf( stderr, "Can't open configuration file '%s'\n", config_file );
    exit( 1 );
  }

  fprintf( file_out, "Configuration file: '%s'\n", config_file );
  fprintf( file_out, "Log file          : '%s'\n\n", log_file );

  /*  get filename of the network  */
  skipComments( file_in );
  ret = fscanf( file_in, "%s", net_file );
  if (ret != 1)
    {
    fprintf( stderr, "Missing name of the networkfile\n" );
    exit( 1 );
  }

  ret = krui_loadNet( net_file, &netname );
  errChk( ret );
  fprintf( file_out, "Networkfile '%s' loaded.\n", net_file );

  krui_getNetInfo( &no_of_sites, &no_of_links, &ret, &ret );
  no_of_units = krui_getNoOfUnits();
  fprintf( file_out, "Network name: %s\n", netname );
  fprintf( file_out, "No. of units: %d\n", no_of_units );
  fprintf( file_out, "No. of sites: %d\n", no_of_sites );
  fprintf( file_out, "No. of links: %d\n", no_of_links );
  fprintf( file_out, "Learning Function: %s\n\n", krui_getLearnFunc() );

  /*  get filename of the pattern file	*/
  skipComments( file_in );
  ret = fscanf( file_in, "%s", pat_file );
  if (ret != 1)
    {
    fprintf( stderr, "Missing name of the patternfile\n" );
    exit( 1 );
  }

  ret = krui_loadPatterns( pat_file );
  errChk( ret );
  fprintf( file_out, "Patternfile '%s' loaded.\n", pat_file);
  no_of_patterns = krui_getNoOfPatterns();
  fprintf( file_out, "No. of patterns: %d\n\n", no_of_patterns );

  /*  get max. number of cycles  */
  skipComments( file_in );
  ret = fscanf( file_in, "%d", &no_of_cycles );
  if (ret != 1)
    {  fprintf( stderr, "Missing number of cycles to learn\n" );
       exit( 1 );
  }
  if (no_of_cycles <= 0)
    {  fprintf( stderr, "Invalid number of cycles to learn\n" );
       exit( 1 );
  }

  fprintf( file_out, "No. of cycles: %d\n\n", no_of_cycles );

  /*  get max. Network Error to stop  */
  skipComments( file_in );
  ret = fscanf( file_in, "%f", &best_net_error );
  if (ret != 1)
    {  fprintf( stderr, "Missing max. network error to stop\n" );
       exit( 1 );
  }
  fprintf( file_out, "Max. network error to stop: %f\n\n", best_net_error );


  /*  get learning parameters  */
  for (i = 0; i < NO_OF_LEARN_PARAMS; i++)
    {
    skipComments( file_in );
    ret = fscanf( file_in, "%f", learn_parameters + i );
    if (ret != 1)
      {
      fprintf( stderr, "Missing learning parameter #%i\n", i );
      exit( 1 );
    }
    fprintf( file_out, "Learning Parameter #%d: %f\n", i + 1, learn_parameters[ i ] );
  }
  fprintf( file_out, "\n" );


  /*  get shuffleing of patterns enable switch	*/
  shuffle = getShuffleingSwitch( file_in );

  fclose( file_in );

  fprintf( file_out, "\n********************************************************\n" );
  clock = 1;
  (void) time( &clock);
  starttime = clock;
  fprintf( file_out, "\n%s Batchlearning started at %s\n",
	   krui_getVersion(), ctime( &clock) );

  /*  create names for temporary files	*/
  tmp_file1 = tempnam( "./", "SNNS_" );
  tmp_file2 = tempnam( "./", "SNNS_" );

  /*  catch signals  */
  catchSignals();

  /*  init cpu time  */
  (void) times( &tp );
  lasttime = tp.tms_utime;

  step = ((no_of_cycles - 1) / NO_OF_CYCLES_DISPLAYED) + 1;

  do
    {
    if (shuffle)
      {  /*  shuffle patterns every cycle  */
      ret = krui_shufflePatterns( TRUE );
      errChk( ret );
    }

/*  REMEMBER:  return_values[ 0 ] returns the current net error
               learn_parameters[ 0 ] contains the 1st learning parameter
*/

    ret = krui_learnAllPatterns( learn_parameters, NO_OF_LEARN_PARAMS, &return_values, &no_of_return_values );
    errChk( ret );
    cycle++;

    /*	print the return values of the learning function  */
    if ((cycle % step == 0) || cycle == (no_of_cycles - 1))
      writeLearnFuncValues();

    /*	get user cpu time  */
    (void) times( &tp );

    if ( ((tp.tms_utime - lasttime) / HZ_value) >= SAVE_PERIOD)
      {
      /*  save temporary network  */
      if (++tmp_switch & 1)
	ret = krui_saveNet( tmp_file1, netname );
      else
	ret = krui_saveNet( tmp_file2, netname );
      errChk( ret );

      if (tmp_switch == 1)
	fprintf( file_out, "\n#####  Temporary network file '%s' created.  #####\n\n", tmp_file1 );
      if (tmp_switch == 2)
	fprintf( file_out, "\n#####  Temporary network file '%s' created.  #####\n\n", tmp_file2 );

      lasttime = tp.tms_utime;
    }
  }
  while ((cycle < no_of_cycles) && (return_values[0] > best_net_error));


  /*  save the network	*/
  ret = krui_saveNet( net_file, netname );
  errChk( ret );
  fprintf( file_out, "\nNetwork saved.\n");

  /*  remove temporary files  */
  if (tmp_switch >= 1)
    {  if (unlink( tmp_file1 ) == 0)
	 fprintf( file_out, "\n#####  Temporary network file '%s' removed.  #####\n", tmp_file1 );
       if (tmp_switch > 1)
	 {  if (unlink( tmp_file2 ) == 0)
	      fprintf( file_out, "#####  Temporary network file '%s' removed.  #####\n", tmp_file2 );
       }
  }

  /*  write statistics	*/
  writeStatistics( NORMAL );
  fclose( file_out );

  fprintf( stdout, "\n%s Batchlearning has terminated successfully.\nLogfile '%s' written.\n",
	   krui_getVersion(), log_file );

  return( 0 );
}
